@@ -95,6 +95,9 @@ struct TotemPlaylistPrivate
 	GtkTreePath *tree_path;
 	GtkTreeViewDropPosition drop_pos;
 
+	/* Cursor ref: 0 if the cursor is unbusy; positive numbers indicate the number of nested calls to set_waiting_cursor() */
+	guint cursor_ref;
+
 	/* This is a scratch list for when we're removing files */
 	GList *list;
 	guint current_to_be_removed : 1;
@@ -145,7 +148,7 @@ typedef struct {
 	TotemPlParserType type;
 } PlaylistSaveType;
 
-static PlaylistSaveType save_types [] = {
+static const PlaylistSaveType save_types [] = {
 	{ NULL, NULL, -1 }, /* By extension entry */
 	{ N_("MP3 ShoutCast playlist"), "pls", TOTEM_PL_PARSER_PLS },
 	{ N_("MP3 audio (streamed)"), "m3u", TOTEM_PL_PARSER_M3U },
@@ -216,21 +219,17 @@ totem_playlist_get_toplevel (TotemPlaylist *playlist)
 }
 
 static void
-totem_playlist_set_waiting_cursor (TotemPlaylist *playlist)
+set_waiting_cursor (TotemPlaylist *playlist)
 {
-	GtkWidget *parent;
-
-	parent = GTK_WIDGET (totem_playlist_get_toplevel (playlist));
-	totem_gdk_window_set_waiting_cursor (parent->window);
+	totem_gdk_window_set_waiting_cursor (gtk_widget_get_window (GTK_WIDGET (totem_playlist_get_toplevel (playlist))));
+	playlist->priv->cursor_ref++;
 }
 
 static void
-totem_playlist_unset_waiting_cursor (TotemPlaylist *playlist)
+unset_waiting_cursor (TotemPlaylist *playlist)
 {
-	GtkWidget *parent;
-
-	parent = GTK_WIDGET (totem_playlist_get_toplevel (playlist));
-	gdk_window_set_cursor (parent->window, NULL);
+	if (--playlist->priv->cursor_ref < 1)
+		gdk_window_set_cursor (gtk_widget_get_window (GTK_WIDGET (totem_playlist_get_toplevel (playlist))), NULL);
 }
 
 static void
@@ -347,6 +346,7 @@ totem_playlist_set_current_subtitle (TotemPlaylist *playlist, const char *subtit
 static char *
 totem_playlist_mrl_to_title (const gchar *mrl)
 {
+	GFile *file;
 	char *filename_for_display, *unescaped;
 
 	if (g_str_has_prefix (mrl, "dvd://") != FALSE) {
@@ -355,22 +355,21 @@ totem_playlist_mrl_to_title (const gchar *mrl)
 		return g_strdup_printf (_("Title %d"), (int) g_strtod (mrl + 6, NULL));
 	} else if (g_str_has_prefix (mrl, "dvb://") != FALSE) {
 		/* This is "BBC ONE(BBC)" for "dvb://BBC ONE(BBC)" */
-		unescaped = g_strdup (mrl + 6);
-	} else {
-		GFile *file;
-
-		file = g_file_new_for_uri (mrl);
-		unescaped = g_file_get_basename (file);
-		g_object_unref (file);
+		return g_strdup (mrl + 6);
 	}
 
+	file = g_file_new_for_uri (mrl);
+	unescaped = g_file_get_basename (file);
+	g_object_unref (file);
+
 	filename_for_display = g_filename_to_utf8 (unescaped,
 			-1,             /* length */
 			NULL,           /* bytes_read */
 			NULL,           /* bytes_written */
 			NULL);          /* error */
 
-	if (filename_for_display == NULL) {
+	if (filename_for_display == NULL)
+	{
 		filename_for_display = g_locale_to_utf8 (unescaped,
 				-1, NULL, NULL, NULL);
 		if (filename_for_display == NULL) {
@@ -397,16 +396,33 @@ totem_playlist_update_save_button (TotemPlaylist *playlist)
 	gtk_widget_set_sensitive (button, state);
 }
 
-static void
-totem_playlist_save_get_iter_func (GtkTreeModel *model,
-		GtkTreeIter *iter, char **uri, char **title,
-		gboolean *custom_title, gpointer user_data)
+static gboolean
+totem_playlist_save_iter_foreach (GtkTreeModel *model,
+				  GtkTreePath  *path,
+				  GtkTreeIter  *iter,
+				  gpointer      user_data)
 {
+	TotemPlPlaylist *playlist = user_data;
+	TotemPlPlaylistIter pl_iter;
+	gchar *uri, *title;
+	gboolean custom_title;
+
 	gtk_tree_model_get (model, iter,
-			URI_COL, uri,
-			FILENAME_COL, title,
-			TITLE_CUSTOM_COL, custom_title,
-			-1);
+			    URI_COL, &uri,
+			    FILENAME_COL, &title,
+			    TITLE_CUSTOM_COL, &custom_title,
+			    -1);
+
+	totem_pl_playlist_append (playlist, &pl_iter);
+	totem_pl_playlist_set (playlist, &pl_iter,
+			       TOTEM_PL_PARSER_FIELD_URI, uri,
+			       TOTEM_PL_PARSER_FIELD_TITLE, (custom_title) ? title : NULL,
+			       NULL);
+
+	g_free (uri);
+	g_free (title);
+
+	return FALSE;
 }
 
 void
@@ -418,13 +434,22 @@ totem_playlist_save_current_playlist (TotemPlaylist *playlist, const char *outpu
 void
 totem_playlist_save_current_playlist_ext (TotemPlaylist *playlist, const char *output, TotemPlParserType type)
 {
+	TotemPlPlaylist *pl_playlist;
 	GError *error = NULL;
+	GFile *output_file;
 	gboolean retval;
 
-	retval = totem_pl_parser_write (playlist->priv->parser,
-			playlist->priv->model,
-                        totem_playlist_save_get_iter_func,
-			output, type, NULL, &error);
+	pl_playlist = totem_pl_playlist_new ();
+	output_file = g_file_new_for_commandline_arg (output);
+
+	gtk_tree_model_foreach (playlist->priv->model,
+				totem_playlist_save_iter_foreach,
+				pl_playlist);
+
+	retval = totem_pl_parser_save (playlist->priv->parser,
+				       pl_playlist,
+				       output_file,
+				       NULL, type, &error);
 
 	if (retval == FALSE)
 	{
@@ -432,6 +457,9 @@ totem_playlist_save_current_playlist_ext (TotemPlaylist *playlist, const char *o
 				error->message, playlist);
 		g_error_free (error);
 	}
+
+	g_object_unref (pl_playlist);
+	g_object_unref (output_file);
 }
 
 static void
@@ -456,6 +484,15 @@ gtk_tree_selection_has_selected (GtkTreeSelection *selection)
 	return retval;
 }
 
+static void
+drop_finished_cb (TotemPlaylist *playlist, GAsyncResult *result, gpointer user_data)
+{
+	/* Emit the "changed" signal once the last dropped MRL has been added to the playlist */
+	g_signal_emit (G_OBJECT (playlist),
+	               totem_playlist_table_signals[CHANGED], 0,
+	               NULL);
+}
+
 static void
 drop_cb (GtkWidget        *widget,
          GdkDragContext   *context, 
@@ -481,7 +518,7 @@ drop_cb (GtkWidget        *widget,
 	if (context->action == GDK_ACTION_MOVE)
 		totem_playlist_clear (playlist);
 
-	list = g_uri_list_extract_uris ((char *)data->data);
+	list = g_uri_list_extract_uris ((char *) gtk_selection_data_get_data (data));
 	file_list = NULL;
 
 	for (i = 0; list[i] != NULL; i++) {
@@ -496,8 +533,6 @@ drop_cb (GtkWidget        *widget,
 		return;
 	}
 
-	totem_playlist_set_waiting_cursor (playlist);
-
 	playlist->priv->tree_path = gtk_tree_path_new ();
 	gtk_tree_view_get_dest_row_at_pos (GTK_TREE_VIEW (playlist->priv->treeview),
 					   x, y,
@@ -530,7 +565,12 @@ drop_cb (GtkWidget        *widget,
 			}
 		}
 
-		totem_playlist_add_mrl (playlist, filename, title);
+		/* Add the MRL to the playlist asynchronously. If it's the last MRL, emit the "changed"
+		 * signal once we're done adding it */
+		if (p->next == NULL)
+			totem_playlist_add_mrl (playlist, filename, title, TRUE, NULL, (GAsyncReadyCallback) drop_finished_cb, NULL);
+		else
+			totem_playlist_add_mrl (playlist, filename, title, TRUE, NULL, NULL, NULL);
 
 		g_free (filename);
 	}
@@ -540,12 +580,6 @@ drop_cb (GtkWidget        *widget,
 	gtk_drag_finish (context, TRUE, FALSE, _time);
 	gtk_tree_path_free (playlist->priv->tree_path);
 	playlist->priv->tree_path = NULL;
-
-	totem_playlist_unset_waiting_cursor (playlist);
-
-	g_signal_emit (G_OBJECT (playlist),
-			totem_playlist_table_signals[CHANGED], 0,
-			NULL);
 }
 
 void
@@ -822,22 +856,18 @@ totem_playlist_add_files (GtkWidget *widget, TotemPlaylist *playlist)
 {
 	GSList *filenames, *l;
 
-	filenames = totem_add_files (totem_playlist_get_toplevel (playlist),
-			NULL);
+	filenames = totem_add_files (totem_playlist_get_toplevel (playlist), NULL);
 	if (filenames == NULL)
 		return;
 
-	totem_playlist_set_waiting_cursor (playlist);
-
 	for (l = filenames; l != NULL; l = l->next) {
 		char *mrl;
 
 		mrl = l->data;
-		totem_playlist_add_mrl (playlist, mrl, NULL);
+		totem_playlist_add_mrl (playlist, mrl, NULL, TRUE, NULL, NULL, NULL);
 		g_free (mrl);
 	}
 
-	totem_playlist_unset_waiting_cursor (playlist);
 	g_slist_free (filenames);
 }
 
@@ -1213,22 +1243,22 @@ set_playing_icon (GtkTreeViewColumn *column, GtkCellRenderer *renderer,
 		  GtkTreeModel *model, GtkTreeIter *iter, TotemPlaylist *playlist)
 {
 	TotemPlaylistStatus playing;
-	const char *stock_id;
+	const char *icon_name;
 
 	gtk_tree_model_get (model, iter, PLAYING_COL, &playing, -1);
 
 	switch (playing) {
 		case TOTEM_PLAYLIST_STATUS_PLAYING:
-			stock_id = GTK_STOCK_MEDIA_PLAY;
+			icon_name = "media-playback-start-symbolic";
 			break;
 		case TOTEM_PLAYLIST_STATUS_PAUSED:
-			stock_id = GTK_STOCK_MEDIA_PAUSE;
+			icon_name = "media-playback-pause-symbolic";
 			break;
 		default:
-			stock_id = NULL;
+			icon_name = NULL;
 	}
 
-	g_object_set (renderer, "stock-id", stock_id, NULL);
+	g_object_set (renderer, "icon-name", icon_name, NULL);
 }
 
 static void
@@ -1304,6 +1334,51 @@ treeview_row_changed (GtkTreeView *treeview, GtkTreePath *arg1,
 	}
 }
 
+static gboolean
+search_equal_is_match (const gchar * s, const gchar * lc_key)
+{
+	gboolean match = FALSE;
+
+	if (s != NULL) {
+		gchar *lc_s;
+
+		/* maybe also normalize both strings? */
+		lc_s = g_utf8_strdown (s, -1);
+		match = (lc_s != NULL && strstr (lc_s, lc_key) != NULL);
+		g_free (lc_s);
+	}
+
+	return match;
+}
+
+static gboolean
+search_equal_func (GtkTreeModel *model, gint col, const gchar *key,
+                   GtkTreeIter *iter, gpointer userdata)
+{
+	gboolean match;
+	gchar *lc_key, *fn = NULL;
+
+	lc_key = g_utf8_strdown (key, -1);
+
+        /* type-ahead search: first check display filename / title, then URI */
+	gtk_tree_model_get (model, iter, FILENAME_COL, &fn, -1);
+	match = search_equal_is_match (fn, lc_key);
+	g_free (fn);
+
+	if (!match) {
+		gchar *uri = NULL;
+
+		gtk_tree_model_get (model, iter, URI_COL, &uri, -1);
+		fn = g_filename_from_uri (uri, NULL, NULL);
+		match = search_equal_is_match (fn, lc_key);
+		g_free (fn);
+		g_free (uri);
+	}
+
+	g_free (lc_key);
+	return !match; /* needs to return FALSE if row matches */
+}
+
 static void
 init_treeview (GtkWidget *treeview, TotemPlaylist *playlist)
 {
@@ -1340,6 +1415,10 @@ init_treeview (GtkWidget *treeview, TotemPlaylist *playlist)
 
 	playlist->priv->selection = selection;
 
+	/* make type-ahead search work in the playlist */
+	gtk_tree_view_set_search_equal_func (GTK_TREE_VIEW (treeview),
+	                                     search_equal_func, NULL, NULL);
+
 	gtk_widget_show (treeview);
 }
 
@@ -1744,38 +1823,36 @@ totem_playlist_add_one_mrl (TotemPlaylist *playlist,
 	return TRUE;
 }
 
-gboolean
-totem_playlist_add_mrl_with_cursor (TotemPlaylist *playlist, const char *mrl,
-				    const char *display_name)
+typedef struct {
+	GAsyncReadyCallback callback;
+	gpointer user_data;
+	gboolean cursor;
+	TotemPlaylist *playlist;
+	gchar *mrl;
+	gchar *display_name;
+} AddMrlData;
+
+static void
+add_mrl_data_free (AddMrlData *data)
 {
-	gboolean retval;
-
-	totem_playlist_set_waiting_cursor (playlist);
-	retval = totem_playlist_add_mrl (playlist, mrl, display_name);
-	totem_playlist_unset_waiting_cursor (playlist);
-
-	return retval;
+	g_object_unref (data->playlist);
+	g_free (data->mrl);
+	g_free (data->display_name);
+	g_slice_free (AddMrlData, data);
 }
 
-gboolean
-totem_playlist_add_mrl (TotemPlaylist *playlist, const char *mrl,
-			const char *display_name)
+static gboolean
+handle_parse_result (TotemPlParserResult res, TotemPlaylist *playlist, const gchar *mrl, const gchar *display_name)
 {
-	TotemPlParserResult res;
-
-	g_return_val_if_fail (mrl != NULL, FALSE);
-
-	res = totem_pl_parser_parse (playlist->priv->parser, mrl, FALSE);
-
 	if (res == TOTEM_PL_PARSER_RESULT_UNHANDLED)
 		return totem_playlist_add_one_mrl (playlist, mrl, display_name);
-	if (res == TOTEM_PL_PARSER_RESULT_ERROR)
-	{
+	if (res == TOTEM_PL_PARSER_RESULT_ERROR) {
 		char *msg;
 
-		msg = g_strdup_printf (_("The playlist '%s' could not be parsed, it might be damaged."), display_name ? display_name : mrl);
+		msg = g_strdup_printf (_("The playlist '%s' could not be parsed. It might be damaged."), display_name ? display_name : mrl);
 		totem_playlist_error (_("Playlist error"), msg, playlist);
 		g_free (msg);
+
 		return FALSE;
 	}
 	if (res == TOTEM_PL_PARSER_RESULT_IGNORED)
@@ -1784,6 +1861,79 @@ totem_playlist_add_mrl (TotemPlaylist *playlist, const char *mrl,
 	return TRUE;
 }
 
+static void
+add_mrl_cb (TotemPlParser *parser, GAsyncResult *result, AddMrlData *data)
+{
+	TotemPlParserResult res;
+	GSimpleAsyncResult *async_result;
+	GError *error = NULL;
+
+	/* Finish parsing the playlist */
+	res = totem_pl_parser_parse_finish (parser, result, &error);
+
+	/* Remove the cursor, if one was set */
+	if (data->cursor)
+		unset_waiting_cursor (data->playlist);
+
+	/* Create an async result which will return the result to the code which called totem_playlist_add_mrl() */
+	if (error != NULL)
+		async_result = g_simple_async_result_new_from_error (G_OBJECT (data->playlist), data->callback, data->user_data, error);
+	else
+		async_result = g_simple_async_result_new (G_OBJECT (data->playlist), data->callback, data->user_data, totem_playlist_add_mrl);
+
+	/* Handle the various return cases from the playlist parser */
+	g_simple_async_result_set_op_res_gboolean (async_result, handle_parse_result (res, data->playlist, data->mrl, data->display_name));
+
+	/* Free the closure's data, now that we're finished with it */
+	add_mrl_data_free (data);
+
+	/* Synchronously call the calling code's callback function (i.e. what was passed to totem_playlist_add_mrl()'s @callback parameter)
+	 * in the main thread to return the result */
+	g_simple_async_result_complete (async_result);
+}
+
+void
+totem_playlist_add_mrl (TotemPlaylist *playlist, const char *mrl, const char *display_name, gboolean cursor,
+                        GCancellable *cancellable, GAsyncReadyCallback callback, gpointer user_data)
+{
+	AddMrlData *data;
+
+	g_return_if_fail (mrl != NULL);
+
+	/* Display a waiting cursor if required */
+	if (cursor)
+		set_waiting_cursor (playlist);
+
+	/* Build the data struct to pass to the callback function */
+	data = g_slice_new (AddMrlData);
+	data->callback = callback;
+	data->user_data = user_data;
+	data->cursor = cursor;
+	data->playlist = g_object_ref (playlist);
+	data->mrl = g_strdup (mrl);
+	data->display_name = g_strdup (display_name);
+
+	/* Start parsing the playlist. Once this is complete, add_mrl_cb() is called, which will interpret the results and call @callback to
+	 * finish the process. */
+	totem_pl_parser_parse_async (playlist->priv->parser, mrl, FALSE, cancellable, (GAsyncReadyCallback) add_mrl_cb, data);
+}
+
+gboolean
+totem_playlist_add_mrl_finish (TotemPlaylist *playlist, GAsyncResult *result)
+{
+	g_assert (g_simple_async_result_get_source_tag (G_SIMPLE_ASYNC_RESULT (result)) == totem_playlist_add_mrl);
+
+	return g_simple_async_result_get_op_res_gboolean (G_SIMPLE_ASYNC_RESULT (result));
+}
+
+gboolean
+totem_playlist_add_mrl_sync (TotemPlaylist *playlist, const char *mrl, const char *display_name)
+{
+	g_return_val_if_fail (mrl != NULL, FALSE);
+
+	return handle_parse_result (totem_pl_parser_parse (playlist->priv->parser, mrl, FALSE), playlist, mrl, display_name);
+}
+
 static gboolean
 totem_playlist_clear_cb (GtkTreeModel *model,
 			 GtkTreePath  *path,
